/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS AS-IS
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */

/* based on the mips/PMAX/kmin_dma.c code
 * Amiga contributors: Markus  Wild
 *                     Niklas  Hallqvist
 *		       Eduardo Horvath
 *                     Michael L. Hitch
 */

#if defined (A3000) + defined (A2091) + defined (GVP11) > 1
#error Only one of A3000, A2091 and GVP11 can be defined
#endif

#if defined (A3000) || defined (GVP11) || defined (A2091)

#include <mach/std_types.h>
#include <sys/types.h>
#include <chips/busses.h>
#include <machine/machparam.h>		/* spl definitions */
#include <device/io_req.h>
#include <device/conf.h>

#ifdef A3000
#include <machine/a3000.h>
#endif
#ifdef GVP11
#include <machine/gvp11.h>
#endif
#ifdef A2091
#include <machine/a2091.h>
#endif
#include <machine/scsi.h>

#ifdef BUFFERED_DMA
/* This is for those people that must DMA into 16-bit memory, although
   it might be interesting for us others as well, as a DMA buffer isn't
   in VM thus not requiring a max DMA size of 4K.  Maybe this speeds up
   disk accesses in spite of the extra copying involved.  Just note:
   DMA buffers are static, i.e. you can't use the memory for anything
   else.  */
/* 8 Megs of DMA buffer !!! */
#define	DMA_MEMORY_BASE		((u_char *) 0x00200000)
/* partition 16-bit memory in 1M chunks */
#define PER_TGT_DMA_SIZE	0x100000
#define PER_TGT_BURST_SIZE	(PER_TGT_DMA_SIZE>>1)
#endif

#define NAMIGA_DMA	1	/* XXX */

/*
 * Definition of the driver for the auto-configuration program.
 */

int	amiga_dma_probe(), amiga_dma_attach(), amiga_dma_intr();


/* a bit ridiculous I know... */
caddr_t	amiga_dma_std[NAMIGA_DMA] = { 0 };
struct	bus_device *amiga_dma_info[NAMIGA_DMA];
struct	bus_driver amiga_dma_driver =
{
  amiga_dma_probe, 0, amiga_dma_attach, 0, amiga_dma_std,
  "amiga_dma", amiga_dma_info
  };

/*
 * Adapt/Probe/Attach functions
 */
int
amiga_dma_probe (xxx, ui)
     struct bus_device *ui;
{
  /* we are only called if ??? hand-filled us in */
  return 1;
}

int
amiga_dma_attach (ui)
     register struct bus_device *ui;
{
  /* we got nothing to say */
}

int
amiga_dma_intr (unit, spllevel, intr)
{
#ifdef A3000
  if (a3000_dmac.int_status & A3000_DMAC_INT_PENDING)
#endif
#ifdef GVP11
  if (gvp11_dmac.control & GVP11_DMAC_INT_PENDING)
#endif
#ifdef A2091
  if (a2091_dmac.int_status & A2091_DMAC_INT_PENDING)
#endif
    amiga_dma_intr_1 (unit, spllevel, intr);
}


/*
 *	Functions to support DMA for the on-board SCSI
 */

#include <scsi/compat_30.h>
#include <scsi/scsi.h>
#include <scsi/scsi_defs.h>
#include <scsi/adapters/scsi_dma.h>
#include <scsi/adapters/scsi_33C93.h>

/* utilities */

#define	u_min(a,b)	(((a) < (b)) ? (a) : (b))

int dma_in_progress = 0;

static inline vm_offset_t
kvtophys (vm_offset_t kv)
{
#ifndef DMA_MEMORY_BASE
  extern pmap_t kernel_pmap;
  extern vm_offset_t pmap_extract ();

  return pmap_extract (kernel_pmap, kv);
#else
  return kv;
#endif
}

#ifdef A3000
static inline void 
a3000_stop_dma ()
{
  if (dma_in_progress)
    {
      a3000_dmac.flush_DMA = 1;
      while (! (a3000_dmac.int_status & A3000_DMAC_DMA_IDLE)) ;
      a3000_dmac.clear_ints = 1;
      a3000_dmac.stop_DMA = 1;
      
      dma_in_progress = 0;
    }
}  

a3000_dma_read (dma_ptr, count)
    vm_offset_t	dma_ptr;
{
  register vm_offset_t phys;

  /* stop DMA engine first */
  a3000_stop_dma ();

  phys = kvtophys (dma_ptr);

  a3000_dmac.control = A3000_DMAC_DIR_READ;
  a3000_dmac.DMA_address = (void *)phys;
  a3000_dmac.start_DMA = 1;
  
  dma_in_progress = 1;
}

a3000_dma_write (dma_ptr, count)
    vm_offset_t	dma_ptr;
{
  register vm_offset_t phys;

  /* stop DMA engine first */
  a3000_stop_dma ();

  phys = kvtophys (dma_ptr);

  a3000_dmac.control = A3000_DMAC_DIR_WRITE;
  a3000_dmac.DMA_address = (void *)phys;
  a3000_dmac.start_DMA = 1;
  
  dma_in_progress = 1;
}

/*
 * Cold, probe time init
 */
opaque_t
a3000_dma_init (unit, base, dma_bsizep)
    int		unit;
    vm_offset_t	base;
    int		*dma_bsizep;
{
  a3000_dmac.DACK_width = A3000_DMAC_WIDTH;
  return (opaque_t) 0;
}

#define amiga_stop_dma a3000_stop_dma
#define amiga_dma_read a3000_dma_read
#define amiga_dma_write a3000_dma_write
#define amiga_dma_init a3000_dma_init

#endif

#ifdef A2091
static char boyer = 0; /* boyer dma chip needs longword alignment */

static inline void 
a2091_stop_dma ()
{
  if (dma_in_progress)
    {
      a2091_dmac.flush_DMA = 1;
      while (! (a2091_dmac.int_status & A2091_DMAC_DMA_IDLE)) ;
      a2091_dmac.clear_ints = 1;
      a2091_dmac.stop_DMA = 1;
      
      dma_in_progress = 0;
    }
}  

a2091_dma_read (dma_ptr, count)
     vm_offset_t	dma_ptr;
     u_long		count;
{
  register vm_offset_t phys;

  /* stop DMA engine first */
  a2091_stop_dma ();

  phys = kvtophys (dma_ptr);
  
  if( boyer )
    phys = phys & (~0x2);

  a2091_dmac.DACK_width = A2091_DMAC_WIDTH; /* prolly not needed */

  a2091_dmac.control = A2091_DMAC_DIR_READ;
  a2091_dmac.word_Xfer_count = count>>1; /* Just in case, some we ignore */
  a2091_dmac.DMA_address = (char *)phys;
printf("\nDMA read to vir %lx pys %lx\n", dma_ptr, phys);
  a2091_dmac.start_DMA = 1;
  dma_in_progress = 1;
}

a2091_dma_write (dma_ptr, count)
    vm_offset_t	dma_ptr;
{
  register vm_offset_t phys;

  /* stop DMA engine first */
  a2091_stop_dma ();

  phys = kvtophys (dma_ptr);
  if( boyer )
    phys = phys & (~0x2);

  a2091_dmac.DACK_width = A2091_DMAC_WIDTH; /* prolly not needed */

  a2091_dmac.control = A2091_DMAC_DIR_WRITE;
  a2091_dmac.word_Xfer_count = count>>1; /* Just in case, some we ignore */
  a2091_dmac.DMA_address = (void *)phys;
printf("\nDMA write from vir %lx pys %lx\n", dma_ptr, phys);
  a2091_dmac.start_DMA = 1;
  dma_in_progress = 1;
}

/*
 * Cold, probe time init
 */
opaque_t
a2091_dma_init (unit, base, dma_bsizep)
    int		unit;
    vm_offset_t	base;
    int		*dma_bsizep;
{
  int i;
  u_short testpat[] =  { 0x0, 0xffff, 0x5555 };
  
  a2091_dmac.DACK_width = A2091_DMAC_WIDTH;

  /* Test for Boyer DMA chip */

  for( i=0; i<3; i++ ) {
    ((volatile u_short*)&a2091_dmac.word_Xfer_count)[1] = testpat[i];
    if( ((volatile u_short*)&a2091_dmac.word_Xfer_count)[1] != testpat[i] )
      {
	boyer = 0;
	return (opaque_t) 0;
      }
  }
  boyer = 1;
  return (opaque_t) 0;
}

#define amiga_stop_dma a2091_stop_dma
#define amiga_dma_read a2091_dma_read
#define amiga_dma_write a2091_dma_write
#define amiga_dma_init a2091_dma_init

void a2091_reset_scsibus(regs, timeout)
{
  int i;

  a2091_dmac.control = PREST;
  
  for(i=0; i<100000; i++);

  a2091_dmac.control = A2091_DMAC_DIR_READ;
}
#endif

#ifdef GVP11
static inline void 
gvp11_stop_dma ()
{
  if (dma_in_progress)
    {
      gvp11_dmac.stop_DMA = 1;
      dma_in_progress = 0;
    }
}  

gvp11_dma_read (dma_ptr, count)
    vm_offset_t	dma_ptr;
{
  register vm_offset_t phys;

  /* stop DMA engine first */
  gvp11_stop_dma ();

  phys = kvtophys (dma_ptr);

  gvp11_dmac.control &= ~GVP11_DMAC_DIR_WRITE;
  gvp11_dmac.DMA_address = (void *)phys;
  gvp11_dmac.start_DMA = 1;
  
  dma_in_progress = 1;
}

gvp11_dma_write (dma_ptr, count)
    vm_offset_t	dma_ptr;
{
  register vm_offset_t phys;

  /* stop DMA engine first */
  gvp11_stop_dma ();

  phys = kvtophys (dma_ptr);

  gvp11_dmac.control |= GVP11_DMAC_DIR_WRITE;
  gvp11_dmac.DMA_address = (void *)phys;
  gvp11_dmac.start_DMA = 1;
  
  dma_in_progress = 1;
}

/*
 * Cold, probe time init
 */
opaque_t
gvp11_dma_init (unit, base, dma_bsizep)
    int		unit;
    vm_offset_t	base;
    int		*dma_bsizep;
{
  gvp11_dmac.secret2 = 1;
  gvp11_dmac.secret1 = 0;
  gvp11_dmac.secret3 = 15;
  while (gvp11_dmac.control & GVP11_DMAC_BUSY)
    ;
  gvp11_dmac.control |= GVP11_DMAC_INT_ENABLE;
  return (opaque_t) 0;
}

#define amiga_stop_dma gvp11_stop_dma
#define amiga_dma_read gvp11_dma_read
#define amiga_dma_write gvp11_dma_write
#define amiga_dma_init gvp11_dma_init

#endif

int
amiga_dma_intr_1 (unit, spllevel, intr)
{
#if 0
printf("Stopping DMA Interrupt\n");
#endif
  amiga_stop_dma ();
}

/*
 * A target exists
 */
void
amiga_dma_new_target (dma, tgt)
    opaque_t	dma;
    target_info_t	*tgt;
{
  /* provide enough space here, the cmd_buf is used for both, commands and
     (small) results of maintanance commands (ie. inquiry)
     MAKE SURE THE BUFFERS ARE LONG ALIGNED!! */
#ifndef DMA_MEMORY_BASE
  static u_int cmds[NSCSI][8][256/4];

  tgt->cmd_ptr = (u_char *) cmds[tgt->masterno][tgt->target_id];
#else
  tgt->cmd_ptr = DMA_MEMORY_BASE + tgt->masterno * PER_TGT_DMA_SIZE * 8
    + tgt->target_id * PER_TGT_DMA_SIZE;
#endif
}

/*
 * Map virtual addresses onto physical ones
 */
void
amiga_dma_map (dma, tgt)
    opaque_t	dma;
    target_info_t	*tgt;
{
  /*
   * We do real DMA unless we do DMA from/to a buffer (not yet implemented
   * for write operations).
   */

  if ((tgt->cur_cmd == SCSI_CMD_WRITE) ||
      (tgt->cur_cmd == SCSI_CMD_LONG_WRITE))
    {
      io_req_t	   ior = tgt->ior;
      register int len = ior->io_count;

      printf ("SCSI ABOUT TO WRITE TO DISK!!\n");
      gimmeabreak();

      tgt->transient_state.out_count = len;

#ifdef DMA_MEMORY_BASE
      /* copy initial data to DMA buffer */
#endif

      /* avoid leaks ?? */
      if (len < tgt->block_size) 
	{
/*	  bzero(to + len, tgt->block_size - len);*/
	  len = tgt->block_size;
	  tgt->transient_state.out_count = len;
	}

      /* see start_cmd & start_dataout */
      tgt->transient_state.dma_offset = 1;
    } 
  else 
    {
      tgt->transient_state.out_count = 0;
      tgt->transient_state.dma_offset = 0;
    }
}

int
amiga_dma_start_cmd (dma, tgt)
    opaque_t	         dma;
    register target_info_t *tgt;
{
  /* contrary to the 3min, this command already starts DMA for the data
     transfer, since there's no CDB to be transmitted */

  if (tgt->transient_state.in_count)
    return amiga_dma_start_datain (dma, tgt);
  else if (tgt->transient_state.out_count)
    return amiga_dma_start_dataout (dma, tgt);
  else
    return 0;
}

void
amiga_dma_end_cmd(dma, tgt, ior)
    opaque_t	  dma;
    target_info_t *tgt;
    io_req_t	  ior;
{
  tgt->dma_ptr = 0;
  tgt->transient_state.dma_offset = 0;
}

void
amiga_dma_end_xfer (dma, tgt, bytes_read)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  bytes_read;
{
  register volatile char  *ptr;
  register int		  nb;

  amiga_stop_dma ();

  /* On read, flush cache (necessary with our DMAC??) XXXX */
  if (bytes_read != 0)
    {
      /* would be nice to do this in parallel with xfer, earlier on */
      ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;
#ifdef DMA_MEMORY_BASE
      /* need to transfer data from DMA buffer to user buffer*/
      if (tgt->cur_cmd == SCSI_CMD_READ ||
          tgt->cur_cmd == SCSI_CMD_LONG_READ) {
              bcopy (ptr, tgt->ior->io_data + tgt->transient_state.copy_count, bytes_read);
      }
#endif
      flush_the_damn_cache (ptr, bytes_read);
    }
}

int
amiga_dma_start_datain (dma, tgt)
    opaque_t		   dma;
    register target_info_t *tgt;
{
  register char	*dma_ptr;
  register int	count;

  /* setup tgt->dma_ptr to ior->io_data, flush cache  */
#ifndef DMA_MEMORY_BASE
  if (tgt->cur_cmd == SCSI_CMD_READ || tgt->cur_cmd == SCSI_CMD_LONG_READ)
    dma_ptr = tgt->ior->io_data;
  else
#endif
    /* swear it won't disconnect midway a xfer or else */
    dma_ptr = tgt->cmd_ptr;

  if (!(tgt->dma_ptr = dma_ptr))
    panic ("amiga_dma_start_datain, no dma_ptr");

  count = tgt->transient_state.in_count;
  if (count <= 0)
    panic ("amiga_dma_start_datain");

  amiga_dma_read(dma_ptr, count);

#ifdef PER_TGT_BURST_SIZE
  count = u_min(count, PER_TGT_BURST_SIZE);
#else
  /* too bad the Amiga DMA can't chain DMA requests ;-( */
  count = u_min (count, PMAP_PGBYTES);
#endif

  return count;
}

int
amiga_dma_restart_datain_1 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  register char *dma_ptr;
  register int  count;

  /* This is the unlikely case we disconnected before
     actually starting the data xfer */
  if (tgt->dma_ptr == 0)
    return amiga_dma_start_datain (dma, tgt);

  /* keep dma_offset anyways */
  dma_ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;

  count = tgt->transient_state.in_count;

  amiga_dma_read (dma_ptr, count);

#ifdef PER_TGT_BURST_SIZE
  /* copy any data read from DMA buffer to user buffer */
  count = u_min(count, PER_TGT_BURST_SIZE);
#else
  /* too bad the Amiga DMA can't chain DMA requests ;-( */
  count = u_min (count, PMAP_PGBYTES);
#endif
  return count;
}

int
amiga_dma_restart_datain_2 (dma, tgt, xferred)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  xferred;
{
/*XXX*/	tgt->transient_state.copy_count = tgt->transient_state.dma_offset;
  tgt->transient_state.dma_offset += xferred;

  return amiga_dma_restart_datain_1 (dma, tgt);
}

void
amiga_dma_restart_datain_3 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  register volatile char  *ptr;
  int			  len;

  /* avoid doing it again!! XXXX */
  ptr = tgt->dma_ptr + tgt->transient_state.copy_count;
  len = tgt->transient_state.dma_offset + tgt->transient_state.copy_count;
  flush_the_damn_cache (ptr, len);
}


/* boolean_t */
int
amiga_dma_start_dataout (dma, tgt, regp, cmd)
    opaque_t		   dma;
    register target_info_t *tgt;
    volatile char	   *regp, cmd;
{
  register char	*dma_ptr;

  /* see dma_map, we have set dma_offset to 1 if transferring from ior, else
     we're transferring from the command buffer */
#ifndef DMA_MEMORY_BASE
  if (tgt->transient_state.dma_offset)
    dma_ptr = tgt->ior->io_data;
  else
#endif
    dma_ptr = tgt->cmd_ptr + tgt->transient_state.cmd_count;

  if (! dma_ptr)
    panic ("amiga_dma_start_dataout no data_ptr!");

  amiga_dma_write (dma_ptr, tgt->transient_state.out_count);

  tgt->transient_state.dma_offset = 0;
  tgt->dma_ptr = dma_ptr;

/*  return TRUE;*/
  /* since we're using start_dataout as part of start_cmd, return the
     transfer length the SBIC should be loaded with */
  return tgt->transient_state.out_count;
}

int
amiga_dma_restart_dataout_1 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  int	count;

  /* this is if we disconnected before starting any xfer */
  if (tgt->dma_ptr == 0)
    amiga_dma_start_dataout (dma, tgt, &count, 0);

  count = tgt->transient_state.out_count;
#ifdef PER_TGT_BURST_SIZE
  count = u_min(count, PER_TGT_BURST_SIZE);
  /* copy user data to DMA buffer */
#else
  /* too bad the Amiga DMA can't chain DMA requests ;-( */
  count = u_min (count, PMAP_PGBYTES);
#endif

  return count;
}

int
amiga_dma_restart_dataout_2 (dma, tgt, xferred)
    opaque_t		   dma;
    register target_info_t *tgt;
    int			   xferred;
{
  tgt->transient_state.dma_offset += xferred;

  return amiga_dma_restart_dataout_1 (dma, tgt);
}

int
amiga_dma_restart_dataout_3 (dma, tgt, regp)
    opaque_t	  dma;
    target_info_t *tgt;
    volatile char *regp;
{
  register char	*dma_ptr;
  int		ret;

  /* keep dma_offset anyways */
  dma_ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;

/* unaligned issues here, align to 8 bytes ? (maybe not..) XXXX */

  amiga_dma_write (dma_ptr, tgt->transient_state.out_count);

  return 0; /* or 1 if adjusted */
}

void
amiga_dma_restart_dataout_4 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* Nothing needed here */
}

int
amiga_dma_start_msgin (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* SBIC ONLY SUPPORTS DATA DMA, NO MSGS */
  return 0;
}

void
amiga_dma_end_msgin (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* SBIC ONLY SUPPORTS DATA DMA, NO MSGS */
}


boolean_t
amiga_dma_disconn_1 (dma, tgt, xferred)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  xferred;
{
  /*
   * XXX Known bug: if we disconnect on an odd bb ??
   */
  amiga_stop_dma ();

  tgt->transient_state.dma_offset += xferred;
  return FALSE;
}

boolean_t
amiga_dma_disconn_2 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* we have to stop DMA here, as the Amiga-DMAC doesn't have a notion
     of a transfer length */
  amiga_stop_dma ();

  return FALSE;
}


boolean_t
amiga_dma_disconn_3 (dma, tgt, xferred)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  xferred;
{
  register volatile char	*ptr;

  amiga_stop_dma ();

  ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;
  flush_the_damn_cache (ptr, xferred);

  tgt->transient_state.dma_offset += xferred;
  return FALSE;
}

#define	amiga_dma_disconn_4		amiga_dma_disconn_3
#define	amiga_dma_disconn_5		amiga_dma_disconn_1
#define	amiga_dma_disconn_callback	0	/* never called */

flush_the_damn_cache (p, nb)
    vm_offset_t p;
{
  vm_size_t off;

  off = p & (PMAP_PGBYTES-1);
  off = PMAP_PGBYTES - off;
  while (nb) 
    {
      off = (nb < off) ? nb : off;
#if 0
      mipscache_Dflush( PHYS_TO_K0SEG(kvtophys(p)), off);
#else
      /* XXXXX HOW DO YOU DO THIS WITH M68K PMAP CODE??? 
         moveq #??,d0
         movec d0,cacr
         ??????? */
#endif
      p += off;
      nb -= off;
      off = PMAP_PGBYTES;
    }
}

scsi_dma_ops_t amiga_dma_ops = {
  amiga_dma_init,
  amiga_dma_new_target,
  amiga_dma_map,
  amiga_dma_start_cmd,
  amiga_dma_end_xfer,
  amiga_dma_end_cmd,
  amiga_dma_start_datain,
  amiga_dma_start_msgin,
  amiga_dma_end_msgin,
  amiga_dma_start_dataout,
  amiga_dma_restart_datain_1,
  amiga_dma_restart_datain_2,
  amiga_dma_restart_datain_3,
  amiga_dma_restart_dataout_1,
  amiga_dma_restart_dataout_2,
  amiga_dma_restart_dataout_3,
  amiga_dma_restart_dataout_4,
  amiga_dma_disconn_1,
  amiga_dma_disconn_2,
  amiga_dma_disconn_3,
  amiga_dma_disconn_4,
  amiga_dma_disconn_5,
  amiga_dma_disconn_callback,
};

/* moved here so I don't have to declare amiga_dma_ops in the header... */
#ifdef A3000
a3000_init ()
{
  a3000_dmac.DACK_width = A3000_DMAC_WIDTH;
  sbic_set_dmaops (0, &amiga_dma_ops);
}
#endif

#ifdef GVP11
gvp11_init ()
{
  sbic_set_dmaops (0, &amiga_dma_ops);
}

#endif

/* moved here so I don't have to declare amiga_dma_ops in the header... */
#ifdef A2091
a2091_init ()
{
  a2091_dmac.DACK_width = A2091_DMAC_WIDTH;
  sbic_set_dmaops (0, &amiga_dma_ops);
}
#endif

#endif
